@@ -31,10 +31,119 @@ from deepcinac.cinac_structures import create_cinac_recording_from_cinac_file_se
 import sklearn.metrics
 
 
+def fusion_cell_type_predictions(cell_type_pred_1, cell_type_pred_2,
+                                 cell_type_config_file=None, filename_to_save=None):
+    """
+    Take two cell type predictions (should be for the same movie, so the same number of cells in both),
+    and fusion it so that cells that have a cell type probabilities in cell_type_pred_2 (sum of probablities equals to 1)
+    will be transfered to
+    cell_type_pred_1, so that the cell type for those cells in cell_type_pred_1 might change.
+    If a yaml cell_type_config_file is provided, then verbose mode will be on and the changes will be print.
+    It is possible to give filename_to_save and the new prediction will be saved in npy format
+    Args:
+        cell_type_pred_1: 2d array (n_cells, n_classes), for each cell give the probability for the cell to be one of
+        the cell type (n_classes represent the number of cell type possible.
+        cell_type_pred_2: 2d array (n_cells, n_classes), for each cell give the probability for the cell to be one of
+        the cell type (n_classes represent the number of cell type possible.
+        cell_type_config_file: (str) yaml cell type that associate a cell type name to a code (one of the column
+        in cell_type_pred)
+        filename_to_save: (str) should have npy extension or it will be added, should include the full path.
+
+    Returns: a 2d array (n_cells, n_classes) representing the new cell type predictions will be returned
+
+    """
+    # first we check the that both cell pred have the same dimensions
+    if cell_type_pred_1.shape[0] != cell_type_pred_2.shape[0]:
+        print(f"Both cell type pred 1 & 2 should have the same number of cells instead we have "
+              f"{cell_type_pred_1.shape[0]} vs {cell_type_pred_2.shape[0]}")
+        return
+
+    if cell_type_pred_1.shape[1] != cell_type_pred_2.shape[1]:
+        print(f"Both cell type pred 1 & 2 should have the same number of cells type instead we have "
+              f"{cell_type_pred_1.shape[1]} vs {cell_type_pred_2.shape[1]}")
+        return
+
+    # not to modify the original one
+    cell_type_pred_1 = np.copy(cell_type_pred_1)
+
+    if cell_type_config_file is not None:
+        cell_type_from_code_dict, cell_type_to_code_dict, multi_class_arg = \
+            read_cell_type_categories_yaml_file(yaml_file=cell_type_config_file)
+        # we display info about the code:
+        print(f"fusion_cell_type_predictions() on {len(cell_type_pred_1)} cells")
+        for code in np.arange(cell_type_pred_1.shape[1]):
+            print(f"Cell type for code {code}: {cell_type_from_code_dict[code]}")
+        print(" ")
+    else:
+        cell_type_from_code_dict = None
+
+    if cell_type_from_code_dict is not None:
+        cell_count_by_type = np.zeros(cell_type_pred_1.shape[1], dtype="int16")
+        for cell in np.arange(len(cell_type_pred_1)):
+            cell_type_code = np.argmax(cell_type_pred_1[cell])
+            cell_count_by_type[cell_type_code] += 1
+        print(" ")
+        print(f"Number of cells for each cell type before fusion:")
+        for code in np.arange(len(cell_type_from_code_dict)):
+            print(f"{cell_count_by_type[code]} {cell_type_from_code_dict[code]}")
+        print(" ")
+
+    n_cells_changed = 0
+    n_cells_different = 0
+
+    for cell in np.arange(len(cell_type_pred_1)):
+        if np.sum(cell_type_pred_2[cell]) == 0:
+            # means no predictions are given for that cell
+            continue
+
+        n_cells_changed += 1
+
+        if cell_type_from_code_dict is not None:
+            # we print some information
+            print(f"Cell type pred of cell {cell} were: {np.round(cell_type_pred_1[cell], 2)} ->"
+                  f"{cell_type_from_code_dict[np.argmax(cell_type_pred_1[cell])]}")
+
+        if np.argmax(cell_type_pred_1[cell]) != np.argmax(cell_type_pred_2[cell]):
+            n_cells_different += 1
+
+        cell_type_pred_1[cell] = cell_type_pred_2[cell]
+        if cell_type_from_code_dict is not None:
+            print(f"New ones are: {np.round(cell_type_pred_1[cell], 2)} ->"
+                  f"{cell_type_from_code_dict[np.argmax(cell_type_pred_1[cell])]}")
+            print(" ")
+
+    if cell_type_from_code_dict is not None:
+        print(f"{n_cells_changed} cells had their predictions changed.")
+        if n_cells_different == 0:
+            print("All cells still have the same cell type.")
+        elif n_cells_different == 1:
+            print("1 cell has now a new cell type.")
+        else:
+            print(f"{n_cells_different} cells have now a news cell type.")
+
+    if cell_type_from_code_dict is not None:
+        cell_count_by_type = np.zeros(cell_type_pred_1.shape[1], dtype="int16")
+        for cell in np.arange(len(cell_type_pred_1)):
+            cell_type_code = np.argmax(cell_type_pred_1[cell])
+            cell_count_by_type[cell_type_code] += 1
+        print(" ")
+        print(f"Number of cells for each cell type after fusion:")
+        for code in np.arange(len(cell_type_from_code_dict)):
+            print(f"{cell_count_by_type[code]} {cell_type_from_code_dict[code]}")
+        print(" ")
+
+    if filename_to_save is not None:
+        if not filename_to_save.endswith(".npy"):
+            filename_to_save = filename_to_save + ".npy"
+        np.save(filename_to_save, cell_type_pred_1)
+
+    return cell_type_pred_1
+
+
 def select_activity_classifier_on_cell_type_outputs(cell_type_config_file, cell_type_predictions,
                                                     cell_type_to_classifier, default_classifier,
                                                     unknown_cell_type_action="ignore",
-                                                    cell_type_threshold=0.5):
+                                                    cell_type_threshold=0.5, verbose=1):
     """
     Allows to create a dictionary that will be passed to CinacPredictor.add_recording() as the model_files_dict
     argument. Each cell will be given a specific activity classifier regarding its cell type.
@@ -53,6 +162,7 @@ def select_activity_classifier_on_cell_type_outputs(cell_type_config_file, cell_
         unknown_cell_type_action: (str) Decide what to do when cell type is not in cell_type_to_classifier. Either
         'ignore' meaning the cell with that cell type won't be predicted, "default" we use the default classifier
         on it or "exception" then an exception would be raise and the excecution will be aborded
+        verbose: (int) if > 0, then the number of cells by cell type will be printed
 
     Returns: return a dict compatible with CinacPredictor.add_recording() model_files_dict argument.
     key is a tuple of 3 strings (model file, weights file, classifier id), and value is the cells whose activity
@@ -76,8 +186,10 @@ def select_activity_classifier_on_cell_type_outputs(cell_type_config_file, cell_
         cell_type_code = cell_type_to_code_dict[cell_type.strip().lower()]
         cell_type_to_classifier_encoded[cell_type_code] = model_tuple
 
+    cell_count_by_type = np.zeros(len(cell_type_from_code_dict), dtype="int16")
     for cell in np.arange(len(cell_type_predictions)):
         cell_type_code = np.argmax(cell_type_predictions[cell])
+        cell_count_by_type[cell_type_code] += 1
         if len(np.where(cell_type_predictions[cell] == cell_type_predictions[cell, cell_type_code])[0]) > 1:
             # mean the cell as two cell types with the same prediction
             if default_classifier not in model_files_dict:
@@ -104,6 +216,14 @@ def select_activity_classifier_on_cell_type_outputs(cell_type_config_file, cell_
     for model_tuple, cells in model_files_dict.items():
         model_files_dict[model_tuple] = np.array(cells)
 
+    if verbose > 0:
+        # print the number of cells for each cell type
+        print(" ")
+        print(f"Number of cells for each cell type:")
+        for code in np.arange(len(cell_type_from_code_dict)):
+            print(f"{cell_count_by_type[code]} {cell_type_from_code_dict[code]}")
+        print(" ")
+
     return model_files_dict
 
 
@@ -167,7 +287,7 @@ class CinacPredictor:
 
     def predict(self, results_path, overlap_value=0.5, output_file_formats="npy", cell_type_classifier_mode=False,
                 n_segments_to_use_for_prediction=2, cell_type_pred_fct=np.mean, create_dir_for_results=True,
-                cell_buffer=None, all_pixels=False,
+                cell_buffer=None, all_pixels=True,
                 **kwargs):
         """
         Will predict the neural activity state of the cell for each frame of the calcium imaging movies.
@@ -178,8 +298,11 @@ class CinacPredictor:
             cell_type_classifier_mode: means we want to classify each cell to know its type. So far return a prediction
             value between 0 and 1, 1 meaning 100% a pyramidal cell, 0 100% sure not (should be interneuron then)
             output_file_formats: a string or list of string, representing the format in which saving the 2d array
-            representing the prediction for each movie. The choices are: "mat" or "npy". If "mat", matlab format, then
-            the predictions named will be "predictions"
+            representing the prediction for each movie. The choices are: "mat" or "npy" or "npz". npy and npz will both
+            comes to a npz format. Two keys will be in the .mat or .npz, one is "predictions" containing a 1d or 2 array
+            with predictions inside (1d array for cell activity predictions, 2d array for cell type activity predicctions
+            (n_cells, n_cells_type), and the second key is "cells" which is a 1d array containing the indices of
+            the cells that has been predicted (other cells will have a 0 value prediction in the predictions array).
             n_segments_to_use_for_prediction: used when cell_type_classifier_mode is at True. Indicate how many segment
             of window_len should be used to predict the cell_type. For example if the value is 2 and window_len is 200,
             then we take the mean of the prediction over 2 segments of 200 frames. Those segments are selected based
@@ -193,7 +316,9 @@ class CinacPredictor:
             **kwargs:
 
 
-        Returns: a dict with key being the id of the cinac recording and value a numpy array of n_cells
+        Returns: a dict with key being the id of the cinac recording and value:
+        numpy array of n_cells containing the predictions (1d for cell activity, and 2d for cell type
+        (n_cells, n_cell_types),.
 
         """
 
@@ -336,7 +461,10 @@ class CinacPredictor:
                 else:
                     new_dir_results = results_path
 
-                file_name = f"{cinac_recording.identifier}_predictions{network_identifier}"
+                predictions_str = "predictions"
+                if cell_type_classifier_mode:
+                    predictions_str = "cell_type_predictions"
+                file_name = f"{cinac_recording.identifier}_{predictions_str}{network_identifier}"
                 for output_file_format in output_file_formats:
                     if "mat" in output_file_format:
                         sio.savemat(os.path.join(new_dir_results, file_name + ".mat"),
